/*
 * input.cpp
 *
 *  Created on: 2021年12月4日
 *      Author: pengzc
 */

#include "input.h"

#include <fcntl.h>
#include <string.h>
#include <unistd.h>
#include <cmath>

#include "utils/Log.h"

#define VTS_MAGIC           0xDEADBEEF
#define VTS_MAX_POINT       10
#define ZKBIT(n)            (1 << (n))

struct zk_vts_data {
    unsigned int package_magic;
    /**
     * u32 package_info:
     * BIT0-BIT3: point个数,最大10点
     * BIT4-BIT13：对应point按下弹起状态，=0弹起，=1按下
     * BIT14-BIT15：=0ABS值，=1KEY值，=2REL值
     */
    unsigned int package_info;
    unsigned int data[VTS_MAX_POINT];
};

static int inject_event(int fd, const zk_vts_point* event, int event_count) {
  if (fd < 0) {
    LOGE("invalid file descriptor");
    return -1;
  }

  if (event_count > VTS_MAX_POINT) {
    LOGE("多点触摸最大支持10点");
    return -1;
  }

  zk_vts_data ts = {0};
  ts.package_magic = VTS_MAGIC;
  ts.package_info |= event_count;
  for (int i = 0; i < event_count; ++i) {
    if (event[i].action == VTS_ACTION_PRESSED) {
      ts.package_info |= (1 << (4 + i));
    }
    //高16位x坐标，低16位y坐标
    ts.data[i] = (event[i].x << 16) | (event[i].y & 0xffff);
  }
  if (write(fd, &ts, int(sizeof(struct zk_vts_data))) < 0) {
      LOGE("fd %d write vts data error %s!\n", fd, strerror(errno));
      return -1;
  }
  return 0;
}

/**
 * 按下
 */
static int inject_event_down(int fd, int x, int y) {
  //down
  zk_vts_point event;
  event.x = x;
  event.y = y;
  event.action = VTS_ACTION_PRESSED;
  CHECK_FAILURE(inject_event(fd, &event, 1));
  return 0;
}

//移动
static int inject_event_move(int fd, int x, int y) {
  zk_vts_point event;
  event.x = x;
  event.y = y;
  event.action = VTS_ACTION_PRESSED;
  CHECK_FAILURE(inject_event(fd, &event, 1));
  return 0;
}

inline int distance(int x1, int y1, int x2, int y2) {
  return sqrt((x2 - x1) * (x2 - x1) + (y2 - y1) * (y2 - y1));
}

static int do_inject_event_move(int fd, int x1, int y1, int x2, int y2) {
  if (distance(x1, y1, x2, y2) < 5) {
    return 0;
  }
  int cx = (x2 + x1) / 2;
  int cy = (y2 + y1) / 2;
  CHECK_FAILURE(do_inject_event_move(fd, x1, y1, cx, cy));
  CHECK_FAILURE(inject_event_move(fd, cx, cy));
  usleep(1000 * Input::kMoveDelayTime);
  CHECK_FAILURE(do_inject_event_move(fd, cx, cy, x2, y2));
  return 0;
}

//抬起
static int inject_event_up(int fd, int x, int y) {
  zk_vts_point event;
  event.x = x;
  event.y = y;
  event.action = VTS_ACTION_RELEASE;
  CHECK_FAILURE(inject_event(fd, &event, 1));
  return 0;
}

Input::Input(const std::string& dev) {
  dev_ = dev;
  fd_ = open(dev.c_str(), O_WRONLY);
  if (fd_ < 0) {
    LOGE("failed to open %s, %s", dev.c_str(), strerror(errno));
  }
}

Input::~Input() {
  if (fd_ >= 0) {
    close(fd_);
  }
}



int Input::Tap(const std::list<Point>& points, int interval) {
  for (auto it = points.begin(); it != points.end(); ++it) {
    if (interval > 0 && it != points.begin()) {
      usleep(1000 * interval);
    }
    CHECK_FAILURE(Tap(it->x, it->y));
  }
  return 0;
}

int Input::Tap(int x, int y) {
  LOGD("\t tap   (%d,%d)", x, y);
  CHECK_FAILURE(inject_event_down(fd_, x, y));
  CHECK_FAILURE(inject_event_up(fd_, x, y));
  return 0;
}

int Input::Swipe(int x1, int y1, int x2, int y2) {
  LOGD("\t swipe (%d,%d) => (%d,%d)", x1, y1, x2, y2);
  CHECK_FAILURE(inject_event_down(fd_, x1, y1));
  CHECK_FAILURE(do_inject_event_move(fd_, x1, y1, x2, y2));
  CHECK_FAILURE(inject_event_up(fd_, x2, y2));
  return 0;
}


int Input::Inject(const zk_vts_point* event, int event_count) {
  if (fd_ < 0) {
    LOGE("invalid file descriptor");
    return -1;
  }
  if (inject_event(fd_, event, event_count) != 0) {
      return -1;
  }
  return 0;
}
